@@ -1,6 +1,6 @@
 # Net::TFTP.pm
 #
-# Copyright (c) 1998 Graham Barr <gbarr@pobox.com>. All rights reserved.
+# Copyright (c) 1998,2007 Graham Barr <gbarr@pobox.com>. All rights reserved.
 # This program is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 
@@ -10,7 +10,7 @@ use strict;
 use vars qw($VERSION);
 use IO::File;
 
-$VERSION = "0.15"; # $Id: //depot/tftp/TFTP.pm#5 $
+$VERSION = "0.18";
 
 sub RRQ	  () { 01 } # read request
 sub WRQ	  () { 02 } # write request
@@ -25,16 +25,17 @@ sub new {
     my $pkg = shift;
     my $host = shift;
 
-    bless {
-	Debug 		=> 0,   	# Debug off
-	Timeout 	=> 5,   	# resend after 5 seconds
-	Retries 	=> 5,   	# resend max 5 times
-	Port 		=> 69,  	# tftp port number
-	BlockSize 	=> 0,   	# use default blocksize (512)
-	Mode 		=> 'netascii',	# transfer in netascii
-	@_,				# user overrides
-	Host 		=> $host,	# the hostname
-    }, $pkg;
+    	bless {
+		Debug 		=> 0,   	# Debug off
+		Timeout 	=> 5,   	# resend after 5 seconds
+		Retries 	=> 5,   	# resend max 5 times
+		Port 		=> 69,  	# tftp port number
+		BlockSize 	=> 0,   	# use default blocksize (512)
+		IpMode		=> 'v4',	# Operate in IPv6 mode, off by default
+		Mode 		=> 'netascii',	# transfer in netascii
+		@_,				# user overrides
+		Host 		=> $host,	# the hostname
+	}, $pkg;
 }
 
 sub timeout {
@@ -79,6 +80,13 @@ sub host {
     $v
 }
 
+sub ip_mode {
+    my $self = shift;
+    my $v = $self->{'IpMode'};
+    $self->{'IpMode'} = shift if @_;
+    $v
+}
+
 sub ascii {
     $_[0]->mode('netascii');
 }
@@ -121,7 +129,11 @@ sub get {
 
     my $file = $local;
     unless(ref($local)) {
-	unlink($file);
+	my $retval = unlink($file);
+        if ($retval < 1) {
+                $self->{'error'} = "Can not unlink $file: $!";
+                return undef;
+        }
 	$local = IO::File->new($file,O_WRONLY|O_TRUNC|O_CREAT);
     }
 
@@ -145,7 +157,7 @@ sub get {
     $self->{'error'} = $io->error
 	unless(close($io));
 
-    exists $self->{'error'};
+    exists $self->{'error'} ? undef : 1;
 }
 
 sub put {
@@ -162,7 +174,7 @@ sub put {
 	unless(ref($local)) {
 	    unless ($local = IO::File->new($file,O_RDONLY)) {
 		$self->{'error'} = "$file: $!";
-		return 0;
+		return undef;
 	    }
 	}
     }
@@ -192,7 +204,7 @@ sub put {
     $self->{'error'} = $io->error
 	unless(close($io));
 
-    exists $self->{'error'};
+    exists $self->{'error'} ? undef : 1;
 }
 
 package Net::TFTP::IO;
@@ -208,20 +220,47 @@ sub new {
     my $io = $pkg->SUPER::new;
 
     $opts->{'Mode'} = lc($opts->{'Mode'});
+    $opts->{'IpMode'} = lc($opts->{'IpMode'});
     $opts->{'Mode'} = "netascii"
 	unless $opts->{'Mode'} eq "octet";
     $opts->{'ascii'} = lc($opts->{'Mode'}) eq "netascii";
 
     my $host = $opts->{'Host'};
-    my $port = $host =~ s/:(\d+)$// ? $1 : $opts->{'Port'};
+    do {
+            $tftp->{'error'} = "No hostname given";
+            return undef;
+    } unless defined($host);
+
+    ## jjmb - had to make an adjustment here the logic used originally does not work well
+    ##        with IPv6.
+    my $port = undef;
+    if($opts->{'IpMode'} eq "v6") {
+	    require Socket6;
+	    require IO::Socket::INET6;
+    	$port = $opts->{'Port'};
+    } else {
+    	$port = $host =~ s/:(\d+)$// ? $1 : $opts->{'Port'};
+    }
     my $addr = inet_aton($host);
 
-    unless($addr) {
-	$tftp->{'error'} = "Bad hostname '$host'";
-	return undef;
+    ## jjmb - added some logic here for the time being to prevent some errors from showing
+    if($opts->{'IpMode'} eq "v6") {
+            # Skipping validation
+     } else {
+	    unless($addr) {
+		$tftp->{'error'} = "Bad hostname '$host'";
+		return undef;
+	    }
+    }
+
+    ## jjmb - need to construct different objects depending on the IP version used
+    my $sock = undef;
+    if($opts->{'IpMode'} eq "v6") {
+    	$sock = IO::Socket::INET6->new(PeerAddr => $opts->{'Host'}, Port => $opts->{'Port'}, Proto => 'udp');
+    } else {
+    	$sock = IO::Socket::INET->new(Proto => 'udp');
     }
 
-    my $sock = IO::Socket::INET->new(Proto => 'udp');
     my $mode = $opts->{'Mode'};
     my $pkt  = pack("n a* c a* c", $op, $remote, 0, $mode, 0);
 
@@ -243,7 +282,11 @@ sub new {
 	@{$opts}{'obuf','blk','ack'} = ('',0,-1);
     }
 
-    send($sock,$pkt,0,pack_sockaddr_in($port,inet_aton($host)));
+    if($tftp->{'IpMode'} eq "v6") {
+    	send($sock,$pkt,0,Socket6::sockaddr_in6($port,Socket6::inet_pton(AF_INET6,$host)));
+    } else {
+    	send($sock,$pkt,0,pack_sockaddr_in($port,inet_aton($host)));
+    }
     _dumppkt($sock,1,$pkt) if $opts->{'Debug'};
 
     tie *$io, "Net::TFTP::IO",$opts;
@@ -583,7 +626,11 @@ sub _recv {
     # The struct in $peer can be bigger than needed for AF_INET
     # so could contain garbage at the end. unpacking and re-packing
     # will ensure it is zero filled (Thanks TomC)
-    $peer = pack_sockaddr_in(unpack_sockaddr_in($peer));
+    if($self->{'IpMode'} eq "v6") {
+    	$peer = Socket6::pack_sockaddr_in6(Socket6::unpack_sockaddr_in6($peer));
+    } else {
+    	$peer = pack_sockaddr_in(unpack_sockaddr_in($peer));
+    }
 
     $self->{'peer'} ||= $peer; # Remember first peer
 
@@ -776,6 +823,7 @@ to and OPTIONS are the default transfer options. Valid options are
  Port	    Port to send data to                                    69
  Mode	    Mode to transfer data in, "octet" or "netascii"     "netascii"
  BlockSize  Negotiate size of blocks to use in the transfer        512
+ IpMode	    Indicates whether to operate in IPv6 mode		   "v4"
 
 =back
 
@@ -795,7 +843,7 @@ filehandle must be read ASAP as the server will otherwise timeout.
 
 If the LOCAL option is given then it can be a file name or a reference.
 If it is a reference it is assumed to be a reference that is valid as a
-filehandle. C<get> will return I<true> if the transfer is sucessful and
+filehandle. C<get> will return I<true> if the transfer is successful and
 I<undef> otherwise.
 
 Valid filehandles are
@@ -827,8 +875,8 @@ If the LOCAL option is missing the put will return a filehandle. This
 filehandle must be written to ASAP as the server will otherwise timeout.
 
 If the LOCAL option is given then it can be a file name or a reference.
-If it is a reference it is assumed to be a valid filehandle as descibed above.
-C<put> will return I<true> if the transfer is sucessful and I<undef> otherwise.
+If it is a reference it is assumed to be a valid filehandle as described above.
+C<put> will return I<true> if the transfer is successful and I<undef> otherwise.
 
 =item error
 
@@ -852,6 +900,10 @@ Set or get the values for the various options. If an argument is passed
 then a new value is set for that option and the previous value returned.
 If no value is passed then the current value is returned.
 
+=item ip_mode ( [ VALUE ] )
+
+Set or get which verion of IP to use ("v4" or "v6")
+
 =item ascii
 
 =item netascii
@@ -872,12 +924,8 @@ Graham Barr <gbarr@pobox.com>
 
 =head1 COPYRIGHT
 
-Copyright (c) 1998 Graham Barr. All rights reserved.
+Copyright (c) 1998,2007 Graham Barr. All rights reserved.
 This program is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
 
-=for html <hr>
-
-I<$Id: //depot/tftp/TFTP.pm#5 $>
-
 =cut
